from itertools import chain

import metpy.calc as mpcalc
import numpy as np
from metpy.units import units


def showalter_index(pressure, temperature, dewpt):
    """Calculate Showalter Index from pressure temperature and 850 hPa lcl.

    Showalter Index derived from [Galway1956]_:
    SI = T500 - Tp500

    where:
    T500 is the measured temperature at 500 hPa
    Tp500 is the temperature of the lifted parcel at 500 hPa

    Parameters
    ----------

        pressure : `pint.Quantity`
            Atmospheric pressure level(s) of interest, in order from highest
            to lowest pressure

        temperature : `pint.Quantity`
            Parcel temperature for corresponding pressure

        dewpt (:class: `pint.Quantity`):
            Parcel dew point temperatures for corresponding pressure


     Returns
     -------

     `pint.Quantity`
        Showalter index in delta degrees celsius
    """

    # find the measured temperature and dew point temperature at 850 hPa.
    idx850 = np.where(pressure == 850 * units.hPa)
    T850 = temperature[idx850]
    Td850 = dewpt[idx850]

    # find the parcel profile temperature at 500 hPa.
    idx500 = np.where(pressure == 500 * units.hPa)
    Tp500 = temperature[idx500]

    # Calculate lcl at the 850 hPa level
    lcl_calc = mpcalc.lcl(850 * units.hPa, T850[0], Td850[0])
    lcl_calc = lcl_calc[0]

    # Define start and end heights for dry and moist lapse rate calculations
    p_strt = 1000 * units.hPa
    p_end = 500 * units.hPa

    # Calculate parcel temp when raised dry adiabatically from surface to lcl
    dl = mpcalc.dry_lapse(lcl_calc, temperature[0], p_strt)
    dl = (dl.magnitude - 273.15) * units.degC  # Change units to C

    # Calculate parcel temp when raised moist adiabatically from lcl to 500mb
    ml = mpcalc.moist_lapse(p_end, dl, lcl_calc)

    # Calculate the Showalter index
    shox = Tp500 - ml
    return shox


def get_skewt_vars(p, tc, tdc, pro):
    """This function processes the dataset values and returns a string element
    which can be used as a subtitle to replicate the styles of NCL Skew-T
    Diagrams.

    Args:

        p (:class: `pint.quantity.build_quantity_class.<locals>.Quantity`):
            Pressure level input from dataset

        tc (:class: `pint.quantity.build_quantity_class.<locals>.Quantity`):
            Temperature for parcel from dataset

        tdc (:class: `pint.quantity.build_quantity_class.<locals>.Quantity`):
            Dew point temperature for parcel from dataset

        pro (:class: `pint.quantity.build_quantity_class.<locals>.Quantity`):
            Parcel profile temperature converted to degC


    Returns:
        :class: 'str'
    """

    # CAPE
    cape = mpcalc.cape_cin(p, tc, tdc, pro)
    cape = cape[0].magnitude

    # Precipitable Water
    pwat = mpcalc.precipitable_water(p, tdc)
    pwat = (pwat.magnitude / 10) * units.cm  # Convert mm to cm
    pwat = pwat.magnitude

    # Pressure and temperature of lcl
    lcl = mpcalc.lcl(p[0], tc[0], tdc[0])
    plcl = lcl[0].magnitude
    tlcl = lcl[1].magnitude

    # Showalter index
    shox = showalter_index(p, tc, tdc)
    shox = shox[0].magnitude

    # Place calculated values in iterable list
    vals = [plcl, tlcl, shox, pwat, cape]
    vals = np.round(vals).astype(int)

    # Define variable names for calculated values
    names = ['Plcl=', 'Tlcl[C]=', 'Shox=', 'Pwat[cm]=', 'Cape[J]=']

    # Combine the list of values with their corresponding labels
    lst = list(chain.from_iterable(zip(names, vals)))
    lst = map(str, lst)

    # Create one large string for later plotting use
    joined = ' '.join(lst)

    return joined



import typing

import cf_xarray
import numpy as np
import xarray as xr

xr.set_options(keep_attrs=True)

_FREQUENCIES = {"day", "month", "year", "season"}


def _find_time_invariant_vars(dset, time_coord_name):
    if isinstance(dset, xr.Dataset):
        return [
            v for v in dset.variables if time_coord_name not in dset[v].dims
        ]
    return


def _contains_datetime_like_objects(d_arr):
    """Check if a variable contains datetime like objects (either
    np.datetime64, or cftime.datetime)"""
    return np.issubdtype(
        d_arr.dtype,
        np.datetime64) or xr.core.common.contains_cftime_datetimes(d_arr)


def _validate_freq(freq):
    if freq not in _FREQUENCIES:
        raise ValueError(
            f"{freq} is not among supported frequency aliases={list(_FREQUENCIES)}"
        )


def _get_time_coordinate_info(dset, time_coord_name):
    if time_coord_name:
        time = dset[time_coord_name]
    else:
        time = dset.cf["time"]
        time_coord_name = time.name

    if not _contains_datetime_like_objects(time):
        raise ValueError(
            f"The {time_coord_name} coordinate should be either `np.datetime64` or `cftime.datetime`"
        )

    return time_coord_name


def _setup_clim_anom_input(dset, freq, time_coord_name):
    _validate_freq(freq)

    time_coord_name = _get_time_coordinate_info(dset, time_coord_name)
    time_invariant_vars = _find_time_invariant_vars(dset, time_coord_name)
    if time_invariant_vars:
        data = dset.drop_vars(time_invariant_vars)
    else:
        data = dset
    time_dot_freq = ".".join([time_coord_name, freq])

    return data, time_invariant_vars, time_coord_name, time_dot_freq


def climatology(
        dset: typing.Union[xr.DataArray, xr.Dataset],
        freq: str,
        time_coord_name: str = None) -> typing.Union[xr.DataArray, xr.Dataset]:
    """Compute climatologies for a specified time frequency.

    Parameters
    ----------
    dset : xr.Dataset, xr.DataArray
        The data on which to operate

    freq : str
        Climatology frequency alias. Accepted alias:

            - 'day': for daily climatologies
            - 'month': for monthly climatologies
            - 'year': for annual climatologies
            - 'season': for seasonal climatologies

    time_coord_name: str, Optional
         Name for time coordinate to use

    Returns
    -------
    computed_dset : xr.Dataset, xr.DataArray
       The computed climatology data

    Examples
    --------
    >>> import xarray as xr
    >>> import pandas as pd
    >>> import numpy as np
    >>> import geocat.comp
    >>> dates = pd.date_range(start="2000/01/01", freq="M", periods=24)
    >>> ts = xr.DataArray(np.arange(24).reshape(24, 1, 1), dims=["time", "lat", "lon"], coords={"time": dates})
    >>> ts
    <xarray.DataArray (time: 24, lat: 1, lon: 1)>
    array([[[ 0]],

        [[ 1]],

        [[ 2]],
    ...
        [[21]],

        [[22]],

        [[23]]])
    Coordinates:
    * time     (time) datetime64[ns] 2000-01-31 2000-02-29 ... 2001-12-31
    Dimensions without coordinates: lat, lon
    >>> geocat.comp.climatology(ts, 'year')
    <xarray.DataArray (year: 2, lat: 1, lon: 1)>
    array([[[ 5.5]],

        [[17.5]]])
    Coordinates:
    * year     (year) int64 2000 2001
    Dimensions without coordinates: lat, lon
    >>> geocat.comp.climatology(ts, 'season')
    <xarray.DataArray (season: 4, lat: 1, lon: 1)>
    array([[[10.]],

        [[12.]],

        [[ 9.]],

        [[15.]]])
    Coordinates:
    * season   (season) object 'DJF' 'JJA' 'MAM' 'SON'
    Dimensions without coordinates: lat, lon
    """
    data, time_invariant_vars, time_coord_name, time_dot_freq = _setup_clim_anom_input(
        dset, freq, time_coord_name)

    grouped = data.groupby(time_dot_freq)
    # TODO: Compute weighted climatologies when `time_bounds` are available
    clim = grouped.mean(time_coord_name)
    if time_invariant_vars:
        return xr.concat([dset[time_invariant_vars], clim], dim=time_coord_name)
    else:
        return clim


def anomaly(
        dset: typing.Union[xr.DataArray, xr.Dataset],
        freq: str,
        time_coord_name: str = None) -> typing.Union[xr.DataArray, xr.Dataset]:
    """Compute anomalies for a specified time frequency.

    Parameters
    ----------
    dset : xr.Dataset, xr.DataArray
        The data on which to operate

    freq : str
        Anomaly frequency alias. Accepted alias:

            - 'day': for daily anomalies
            - 'month': for monthly anomalies
            - 'year': for annual anomalies
            - 'season': for seasonal anomalies

    time_coord_name: str, Optional
         Name for time coordinate to use

    Returns
    -------
    computed_dset : xr.Dataset, xr.DataArray
       The computed anomaly data

    Examples
    --------
    >>> import xarray as xr
    >>> import pandas as pd
    >>> import numpy as np
    >>> import geocat.comp
    >>> dates = pd.date_range(start="2000/01/01", freq="M", periods=24)
    >>> ts = xr.DataArray(np.arange(24).reshape(24, 1, 1), dims=["time", "lat", "lon"], coords={"time": dates})
    >>> ts
    <xarray.DataArray (time: 24, lat: 1, lon: 1)>
    array([[[ 0]],

        [[ 1]],

        [[ 2]],

    ...

        [[21]],

        [[22]],

        [[23]]])
    Coordinates:
    * time     (time) datetime64[ns] 2000-01-31 2000-02-29 ... 2001-12-31
    Dimensions without coordinates: lat, lon
    >>> geocat.comp.anomaly(ts, 'season')
    <xarray.DataArray (time: 24, lat: 1, lon: 1)>
    array([[[-10.]],

        [[ -9.]],

        [[ -7.]],

    ...

        [[  6.]],

        [[  7.]],

        [[ 13.]]])
    Coordinates:
    * time     (time) datetime64[ns] 2000-01-31 2000-02-29 ... 2001-12-31
        season   (time) <U3 'DJF' 'DJF' 'MAM' 'MAM' ... 'SON' 'SON' 'SON' 'DJF'
    Dimensions without coordinates: lat, lon
    """

    data, time_invariant_vars, time_coord_name, time_dot_freq = _setup_clim_anom_input(
        dset, freq, time_coord_name)

    clim = climatology(data, freq, time_coord_name)
    anom = data.groupby(time_dot_freq) - clim
    if time_invariant_vars:
        return xr.merge([dset[time_invariant_vars], anom])
    else:
        return anom


def month_to_season(
    dset: typing.Union[xr.Dataset, xr.DataArray],
    season: str,
    time_coord_name: str = None,
) -> typing.Union[xr.Dataset, xr.DataArray]:
    """Computes a user-specified three-month seasonal mean.

    This function takes an xarray dataset containing monthly data spanning years and
    returns a dataset with one sample per year, for a specified three-month season.

    Parameters
    ----------
    dset : xr.Dataset, xr.DataArray
        The data on which to operate
    season : str
        A string representing the season to calculate: e.g., "JFM", "JJA".
        Valid values are:

         - DJF
         - JFM
         - FMA
         - MAM
         - AMJ
         - MJJ
         - JJA
         - JAS
         - ASO
         - SON
         - OND
         - NDJ
    time_coord_name: str, Optional
        Name for time coordinate to use

    Returns
    -------
    computed_dset : xr.Dataset, xr.DataArray
       The computed data

    Notes
    -----
    This function requires the number of months to be a multiple of 12, i.e. full years must be provided.
    Time stamps are centered on the season. For example, seasons='DJF' returns January timestamps.
    If a calculated season's timestamp falls outside the original range of monthly values, then the calculated mean
    is dropped.  For example, if the monthly data's time range is [Jan-2000, Dec-2003] and the season is "DJF", the
    seasonal mean computed from the single month of Dec-2003 is dropped.
    """

    time_coord_name = _get_time_coordinate_info(dset, time_coord_name)
    mod = 12
    if dset[time_coord_name].size % mod != 0:
        raise ValueError(
            f"The {time_coord_name} axis length must be a multiple of {mod}.")

    seasons_pd = {
        "DJF": ("QS-DEC", 1),
        "JFM": ("QS-JAN", 2),
        "FMA": ("QS-FEB", 3),
        "MAM": ("QS-MAR", 4),
        "AMJ": ("QS-APR", 5),
        "MJJ": ("QS-MAY", 6),
        "JJA": ("QS-JUN", 7),
        "JAS": ("QS-JUL", 8),
        "ASO": ("QS-AUG", 9),
        "SON": ("QS-SEP", 10),
        "OND": ("QS-OCT", 11),
        "NDJ": ("QS-NOV", 12),
    }
    try:
        (season_pd, season_sel) = seasons_pd[season]
    except KeyError:
        raise KeyError(
            f"contributed: month_to_season: bad season: SEASON = {season}. Valid seasons include: {list(seasons_pd.keys())}"
        )

    start_date = dset[time_coord_name][0]
    end_date = dset[time_coord_name][-1]

    # Compute the three-month means, moving time labels ahead to the middle
    # month.
    month_offset = "MS"
    dset_seasons = dset.resample({
        time_coord_name: season_pd
    },
                                 loffset=month_offset).mean()

    # Filter just the desired season, and trim to the desired time range.
    compute_dset = dset_seasons.sel({
        time_coord_name: dset_seasons[time_coord_name].dt.month == season_sel
    }).sel({time_coord_name: slice(start_date, end_date)})
    return compute_dset



import numpy as np


def _is_duck_array(value):
    """Returns True when ``value`` is array-like."""
    if isinstance(value, np.ndarray):
        return True
    return (hasattr(value, "ndim") and hasattr(value, "shape") and
            hasattr(value, "dtype") and hasattr(value, "__array_function__") and
            hasattr(value, "__array_ufunc__"))



from typing import Any, Iterable

import dask.array as da
import numpy as np
import xarray as xr


def _get_missing_value(data: xr.DataArray, args: dict) -> Any:
    """Attempts to extract `missing_value` or `_FillValue` from either `data`
    or `dict`. If not found, returns `np.nan`

    Parameters
    ----------
    data : :class:`xarray.DataArray`
        Data which may contain `missing_value` or `_FillValue` attributes.
    args : :class:`dict`
        Dictionary which may contain `missing_value` key.

    Returns
    -------
    missing_value : :class:`Any`
        The `missing_value` representation.
    """
    if "missing_value" in args:
        missing_value = args["missing_value"]
    elif isinstance(data, xr.DataArray):
        if "missing_value" in data.attrs:
            missing_value = data.attrs["missing_value"]
        elif "_FillValue" in data.attrs:
            missing_value = data.attrs["_FillValue"]
        else:
            missing_value = np.nan
    else:
        missing_value = np.nan

    return missing_value


def _unchunk_ifneeded(data: da.Array, axis: int) -> da.Array:
    """Returns `data` unchunked along `axis`.

    Parameters
    ----------
    data : :class:`dask.array.Array`
        Data which may be chunked along `axis`.
    axis : :class:`int`
        Axis number which specifies the axis to unchunk.

    Returns
    -------
        data : :class:`dask.array.Array`
            A dask array which is not chunked along the specified axis.
    """
    if isinstance(data, da.Array):
        shape = data.shape
        chunksize = data.chunksize
        axis = _check_axis(axis, data.ndim)
        if shape[axis] != chunksize[axis]:
            data = data.rechunk({axis: -1})
        return data
    else:
        raise TypeError("data must be a dask array.")


def ndpolyfit(x: Iterable,
              y: Iterable,
              deg: int,
              axis: int = 0,
              **kwargs) -> (xr.DataArray, da.Array):
    """An extension to `numpy.polyfit` function to support multi-dimensional
    arrays, Dask arrays, and missing values.

    Parameters
    ----------

    x : :class:`array_like`
        X-coordinate, an iterable object of shape `(M,)`, `(M, 1)`, or `(1, M)` where `M = y.shape(axis)`. It cannot
        have `nan` or missing values.

    y : :class:`array_like`:
        Y-coordinate, an iterable containing the data. It could be list, `numpy.ndarray`, `xr.DataArray`, Dask array.
        or any Iterable convertible to `numpy.ndarray`. In case of Dask Array, The data could be chunked. It is
        recommended not to chunk along the `axis` provided.

    axis : :class:`int`
        Axis to fit the polynomial to. Default is 0.

    deg : :class:`int`
        Degree of the fitting polynomial

    kwargs : :class:`dict`, `optional`
        Extra parameter controlling the method behavior:

        rcond : :class:`float`, `optional`
            Relative condition number of the fit. Refer to `numpy.polyfit` for further details.

        full : :class:`bool`, `optional`
            Switch determining nature of return value. Refer to `numpy.polyfit` for further details.

        w : :class:`array_like`, `optional`
            Weights applied to the y-coordinates of the sample points. Refer to `numpy.polyfit` for further details.

        cov : :class:`bool`, `optional`
            Determines whether to return the covariance matrix. Refer to `numpy.polyfit` for further details.

        missing_value : :class:`number` or :class:`np.nan`, `optional`
            The value to be treated as missing. Default is `np.nan`

        meta : :class:`bool`, `optional`
            If set to `True` and the input, i.e. `y`, is of type `xr.DataArray`, the attributes associated to the
            input are transferred to the output.

    Returns
    -------
    coefficients : :class:`xarray.DataArray` or :class:`numpy.ndarray`
        An array containing the coefficients of the fitted polynomial.

    Examples
    --------
        * Fitting a line to a one dimensional array:

        >>> import numpy as np
        >>> from geocat.comp.polynomial import ndpolyfit
        >>> x = np.arange(10, dtype=np.float)
        >>> y = 2*x + 3
        >>> p = ndpolyfit(x, y, deg=1)
        >>> print(p)
        <xarray.DataArray (dim_0: 2)>
        array([2., 3.])
        Dimensions without coordinates: dim_0
        Attributes:
            deg:             1
            provided_rcond:  None
            full:            False
            weights:         None
            covariance:      False

        * Fitting a second degree polynomial to a one dimensional array:

        >>> y = 4*x*x + 3*x + 2
        >>> p = ndpolyfit(x, y, deg=2)
        >>> print(p)
        <xarray.DataArray (dim_0: 3)>
        array([4., 3., 2.])
        Dimensions without coordinates: dim_0
        Attributes:
            deg:             2
            provided_rcond:  None
            full:            False
            weights:         None
            covariance:      False

        * Fitting polynomial with missing values: Ordinarily NaN's are treated as missing values.
          In this example let's introduce a different value to indicate missing data.

        >>> # Let's introduce some missing values:
        >>> y[7:] = 999
        >>> p = ndpolyfit(x, y, deg=2)
        >>> print(p)
        <xarray.DataArray (dim_0: 3)>
        array([ 21.15909091, -62.14090909,  20.4       ])
        Dimensions without coordinates: dim_0
        Attributes:
            deg:             2
            provided_rcond:  None
            full:            False
            weights:         None
            covariance:      False
        >>> # As you can see, we got a different coefficients
        >>> # Now let's define 999 as missing value
        >>> p = ndpolyfit(x, y, deg=2, missing_value=999)
        >>> print(p)
        <xarray.DataArray (dim_0: 3)>
        array([4., 3., 2.])
        Dimensions without coordinates: dim_0
        Attributes:
            deg:             2
            provided_rcond:  None
            full:            False
            weights:         None
            covariance:      False
        >>> # Now we got the coefficient we were looking for

        * Fitting polynomial with NaN as missing values: NaN is by default considered a missing value all the time

        >>> import numpy as np
        >>> from geocat.comp.polynomial import ndpolyfit
        >>> x = np.arange(10, dtype=np.float)
        >>> y = 4*x*x + 3*x + 2
        >>> y[7:] = np.nan
        >>> print(y)
        [  2.   9.  24.  47.  78. 117. 164.  nan  nan  nan]
        >>> p = ndpolyfit(x, y, deg=2)
        >>> print(p)
        <xarray.DataArray (dim_0: 3)>
        array([4., 3., 2.])
        Dimensions without coordinates: dim_0
        Attributes:
            deg:             2
            provided_rcond:  None
            full:            False
            weights:         None
            covariance:      False
        >>> # as you can see, despite not specifying NaN as missing value, the coefficients are properly calculated

        * Fitting a line to a multi-dimensional array

        >>> y_md = np.tile(y.reshape(1, 10, 1, 1), [2, 1, 3, 4])
        >>> y_md.shape
        (2, 10, 3, 4)
        >>> print(y)
        [  2.   9.  24.  47.  78. 117. 164. 219. 282. 353.]
        >>> print(y_md[1, :, 1, 1])
        [  2.   9.  24.  47.  78. 117. 164. 219. 282. 353.]
        >>> p = ndpolyfit(x, y_md, deg=2, axis=1)
    """

    rcond = kwargs.get("rcond", None)
    full = kwargs.get("full", False)
    w = kwargs.get("w", None)
    cov = kwargs.get("cov", False)
    missing_value = _get_missing_value(y, kwargs)
    meta = kwargs.get("meta", True)

    # converting x to numpy.ndarray
    if isinstance(x, xr.DataArray):
        x = x.data
    elif not isinstance(x, np.ndarray):
        x = np.asarray(x)

    if isinstance(y, np.ndarray):
        return _ndpolyfit(x, y, axis, deg, rcond, full, w, cov, missing_value)
    if isinstance(y, xr.DataArray):
        output = _ndpolyfit(x, y.data, axis, deg, rcond, full, w, cov,
                            missing_value)
        attrs = output.attrs

        if meta:
            for attr, v in y.attrs.items():
                attrs[attr] = v
            axis = axis if axis >= 0 else axis + y.ndim
            dims = [("poly_coef" if i == axis else output.dims[i])
                    for i in range(y.ndim)]
            coords = {k: v for k, v in y.coords.items() if k != y.dims[axis]}
            coords["poly_coef"] = list(range(int(deg) + 1))
            output = xr.DataArray(output.data,
                                  attrs=attrs,
                                  dims=dims,
                                  coords=coords)
        return output
    if isinstance(y, da.Array):
        y = _unchunk_ifneeded(y, axis)

        return y.map_blocks(lambda b: _ndpolyfit(x, b, axis, deg, rcond, full,
                                                 w, cov, missing_value, False),
                            dtype=np.float64).compute()
    else:
        return _ndpolyfit(np.asarray(y), x, axis, deg, rcond, full, w, cov,
                          missing_value)


def _ndpolyfit(x: np.ndarray,
               y: np.ndarray,
               axis: int = 0,
               deg: int = 1,
               rcond=None,
               full=False,
               w=None,
               cov=False,
               missing_value=np.nan,
               xarray_output=True) -> (np.ndarray, xr.DataArray):
    """An extension to `numpy.polyfit` function to support multi-dimensional
    arrays, Dask arrays, and missing values.

    Parameters
    ----------

    x : :class:`array_like`
        X-coordinate, an iterable object of shape `(M,)`, `(M, 1)`, or `(1, M)` where `M = y.shape(axis)`.
        It cannot have `nan` or missing values.

    y : :class:`array_like`
        Y-coordinate, an iterable containing the data. It could be list, `numpy.ndarray`, `xr.DataArray`, Dask array.
        or any Iterable convertible to `numpy.ndarray`. In case of Dask Array, The data could be chunked. It is
        recommended not to chunk along the `axis` provided.

    axis : :class:`int`
        Axis to fit the polynomial to. Default is 0.

    deg : :class:`int`
        Degree of the fitting polynomial

    rcond : :class:`float`, `optional`
        Relative condition number of the fit. Refer to `numpy.polyfit` for further details.

    full : :class:`bool`, `optional`
        Switch determining nature of return value. Refer to `numpy.polyfit` for further details.

    w : :class:`array_like`, `optional`
        Weights applied to the y-coordinates of the sample points. Refer to `numpy.polyfit` for further details.

    cov : :class:`bool`, `optional`
        Determines whether to return the covariance matrix. Refer to `numpy.polyfit` for further details.

    missing_value : :class:`number` or :class:`np.nan`, `optional`
        The value to be treated as missing. Default is `np.nan`

    xarray_output : :class:`bool`, `optional`
        Determines the type of the output. If set to `True` the output would be of type `xr.DataArray`
        and the some extra information are attached to the output as attributes. Otherwise, the output
        would be of type `np.ndarray` containing only the coefficients of the fitted polynomial.

    Returns
    -------
    coefficients : :class:`xarray.DataArray` or :class:`numpy.ndarray`
        An array containing the coefficients of the fitted polynomial.
    """

    if not isinstance(x, np.ndarray):
        raise TypeError("x must be a numpy.ndarray")

    if not isinstance(y, np.ndarray):
        raise TypeError(
            "This function only accepts np.ndarray as its input for y")

    axis = _check_axis(axis, y.ndim)

    if x.size != y.shape[axis]:
        raise ValueError(
            "X must have the same number of elements as the y-dimension defined by axis"
        )

    if x.shape not in ((y.shape[axis],), (y.shape[axis], 1), (1,
                                                              y.shape[axis])):
        raise ValueError(
            "x must be of size (M,), (M, 1), or (1, M); where M = y.shape[axis]"
        )

    x = x.reshape(y.shape[axis])

    if deg < 0:
        raise ValueError("deg must be zero or a positive integer.")
    elif int(deg) != deg:
        raise TypeError("deg must be an integral type.")

    deg = int(deg)

    y_rearranged, trailing_shape = _rearrange_axis(y, axis)

    if not np.isnan(missing_value):
        y_rearranged[y_rearranged == missing_value] = np.nan

    mask = np.logical_not(np.isfinite(y_rearranged))

    if mask.any():
        all_row_missing = np.all(mask, axis=1)
        row_with_missing = np.any(mask, axis=1)
        if np.all(all_row_missing == row_with_missing):
            rows_to_keep = np.logical_not(all_row_missing)
            y_rearranged = y_rearranged[rows_to_keep, :]
            x = x[rows_to_keep]
            # mask = np.full(y_rearranged.shape, False, dtype=bool)  # This is
            # not actually needed any longer
            has_missing = False
        else:
            has_missing = True
    else:
        has_missing = False

    if np.isnan(x).any():
        raise ValueError("x cannot have missing values")

    if has_missing:
        tmp_results = []

        for c in range(y_rearranged.shape[1]):
            idx = np.logical_not(mask[:, c])
            tmp_results.append(
                np.polyfit(x[idx],
                           y_rearranged[idx, c],
                           deg=deg,
                           rcond=rcond,
                           full=full,
                           w=w,
                           cov=cov))

        polyfit_output = tmp_results[0].reshape((-1, 1)) if isinstance(
            tmp_results[0], np.ndarray) else tmp_results[0][0].reshape((-1, 1))
        for i in range(1, y_rearranged.shape[1]):
            polyfit_output = np.concatenate(
                (polyfit_output, (tmp_results[c].reshape((-1, 1)) if isinstance(
                    tmp_results[c], np.ndarray) else tmp_results[c][0].reshape(
                        (-1, 1)))),
                axis=1)

    else:
        polyfit_output = np.polyfit(x,
                                    y_rearranged,
                                    deg=deg,
                                    rcond=rcond,
                                    full=full,
                                    w=w,
                                    cov=cov)

    p = _reverse_rearrange_axis((polyfit_output if isinstance(
        polyfit_output, np.ndarray) else polyfit_output[0]), axis,
                                trailing_shape)

    if bool(xarray_output):
        attrs = {
            "deg": deg,
            "provided_rcond": rcond,
            "full": full,
            "weights": w,
            "covariance": cov
        }

        if not has_missing:
            if full:  # full == True
                attrs["residuals"] = polyfit_output[1]
                attrs["rank"] = polyfit_output[2]
                attrs["singular_values"] = polyfit_output[3]
                attrs["rcond"] = polyfit_output[4]
            elif cov:  # (full == False) and (cov == True)
                attrs["V"] = polyfit_output[1]

        return xr.DataArray(p, attrs=attrs)
    else:
        return p


def _check_axis(axis, ndim) -> int:
    if (axis > (ndim - 1)) or (axis < -ndim):
        raise ValueError(
            f"axis must be an integer between {-ndim} and {ndim - 1}")
    if int(axis) != axis:
        raise TypeError("axis must be an integral type.")
    elif axis < 0:
        axis += ndim

    return int(axis)


def _rearrange_axis(data: np.ndarray, axis: int = 0) -> tuple:
    """rearranges the `numpy.ndarray` as a two-dimensional array of size (n,

    -1), where n is the number of elements of the dimension defined by `axis`.

    Parameters
    ----------
    data : :class:`numpy.ndarray`
        An array to be rearranged
    axis : :class:`int`
        The axis that all other dimensions are rearranged around it. default is 0.

    Returns
    -------
    tuple (data :class:`numpy.ndarray`, shape :class:`tuple`):
        A tuple, where the first element contains the reshaped data, and the second is a tuple with all dimensions except the one specified by the axis.
    """
    if not isinstance(data, np.ndarray):
        raise TypeError("data must be a numpy.ndarray.")

    axis = _check_axis(axis, data.ndim)

    if axis != 0:
        data = np.moveaxis(data, axis, 0)

    trailing_shape = data.shape[1:]

    data = data.reshape((data.shape[0], -1))

    return data, trailing_shape


def _reverse_rearrange_axis(data: np.ndarray, axis,
                            trailing_shape: tuple) -> np.ndarray:
    return np.moveaxis(data.reshape((data.shape[0], *trailing_shape)), 0, axis)


def _isvector(input: Iterable) -> bool:
    if isinstance(input, np.ndarray):
        return (input.size != 1) and ((input.ndim == 1) or
                                      ((input.ndim == 2) and
                                       (np.any(input.shape == 1))))
    else:
        return _isvector(np.asarray(input))


def _to_numpy_ndarray(data: Iterable) -> np.ndarray:
    if isinstance(data, xr.DataArray):
        data = data.data
    elif not isinstance(data, np.ndarray):
        data = np.asarray(data)

    return data


def ndpolyval(p: Iterable, x: Iterable, axis: int = 0, **kwargs):
    """Extended version of `numpy.polyval` to support multi-dimensional outputs
    provided by `geocat.comp.ndpolyfit`.

    As the name suggest, this version supports a multi-dimensional `p` array. Let's say `p` is of dimension `(s0,s1,s2)`
    and `axis=1`, then the output would be of dimension `(s0, M, s2)` where M depends on `x`.
    The same way, `x` could be a multi dimensional array or a single array. In another word, `x` is either of
    dimension `(M, )`, `(M, 1)`, `(1, M)` or, in this example, of dimension `(s0, M, s2)`. When `x` is not the vector,
    it must have the same dimension as of `p` except for the dimension that is defined by `axis`.

    Parameters
    ----------

    p : :class:`Iterable`
        Polynomial coeficients.

    x : :class:`Iterable`
        Coordinates where polynomial must be evaluated.

    axis : :class:`int`
        The axis along which to evaluate the polynomial.

    **kwargs:
        Currently not used.

    Returns
    -------
    output : :class:`xr.DataArray`
        Polynomial evaluated with the provided coordinates.

    Examples
    --------

        * Evaluating a polynomial:

        >>> p = [2, 3] # representing y = 2*x + 3
        >>> x = np.arange(-5, 6, dtype="float")
        >>> x
        array([-5., -4., -3., -2., -1.,  0.,  1.,  2.,  3.,  4.,  5.])
        >>> from geocat.comp.polynomial import ndpolyval
        >>> y = ndpolyval(p, x)
        >>> y.shape
        (11,)
        >>> print(y)
        <xarray.DataArray (dim_0: 11)>
        array([-7., -5., -3., -1.,  1.,  3.,  5.,  7.,  9., 11., 13.])
        Dimensions without coordinates: dim_0
        >>> np.testing.assert_almost_equal(y, 2*x+3)

        * evaluating a multi-dimensional fitted polynomial:

        >>> p = np.tile(np.asarray(p).reshape(1, 2, 1, 1), [3, 1, 4, 5])
        >>> p.shape
        (3, 2, 4, 5)
        >>> p[1, :, 1, 1]
        array([2, 3])
        >>> p[1, :, 1, 2]
        array([2, 3])
        >>> y = ndpolyval(p, x, axis=1)
        >>> y.shape
        (3, 11, 4, 5)

        * Fitting a first degree polynomial and calculating the residual manually:

        >>> from geocat.comp.polynomial import ndpolyfit, ndpolyval
        >>> import numpy as np
        >>> n = 10
        >>> x = np.linspace(0, 1, n)
        >>> print(x.shape)
        (10,)
        >>> y = np.random.random((3, 4, n, 2, 5))
        >>> print(y.shape)
        (3, 4, 10, 2, 5)
        >>> # First fitting a first degree polynomial to our 5-Dimensional array
        ... p = ndpolyfit(x, y, deg=1, axis=2)
        >>> print("p.shape: ", p.shape)
        p.shape:  (3, 4, 2, 2, 5)
        >>> # Now re-evaluating the polynomial at the same points:
        ... y_fitted = ndpolyval(p, x, axis=2)
        >>> We are ready to manually calculate the residual now:
        ... signed_residual = y_fitted - y
        >>> y_fitted.shape
        (3, 4, 10, 2, 5)

        * when evaluating x can be another multi-dimensional array

        >>> # Let's work on the p that was calculated in previous example
        ... # using the command:
        ... #           p = ndpolyfit(x, y, deg=1, axis=2)
        ... print("p.shape: ", p.shape)
        p.shape:  (3, 4, 2, 2, 5)

        >>> # Let's pass x which has 20 members this time:
        ... x = np.random.random(20)
        >>> print("x.shape: ", x.shape)
        x.shape:  (20,)
        >>> new_y = ndpolyval(p, x, axis=2)
        >>> print("new_y.shape: ", new_y.shape) # Note new_y.shape[2] = 20
        new_y.shape:  (3, 4, 20, 2, 5)

        >>> # Now let's make a multi-dimensional x
        ... # NOTE: all dimension, except axis=2 is the same as the one in p
        ... #       because during fitting step axis was set to 2
        ... x = np.random.random((3, 4, 42, 2, 5))
        >>> new_y = ndpolyval(p, x, axis=2)
        >>> print("new_y.shape: ", new_y.shape) # Note new_y.shape[2] = 42
        new_y.shape:  (3, 4, 42, 2, 5)

        >>> # now if a wrongly sized array is passed, you would get an error:
        ... x = np.random.random((5, 2, 42, 4, 3))
        >>> new_y = ndpolyval(p, x, axis=2)
        Traceback (most recent call last):
          ...
        ValueError: x has invalid shape.
    """
    p_ndarr = _to_numpy_ndarray(p)
    axis = _check_axis(axis, p_ndarr.ndim)
    if isinstance(x, da.Array):
        x = _unchunk_ifneeded(x, axis)
        x_chunks = list(x.chunks)
        x_chunks[axis] = p_ndarr.shape[axis]
        p_dask = da.from_array(p_ndarr, chunks=x_chunks)
        y = da.map_blocks(
            lambda p_blocks, x_blocks: _ndpolyval(p_blocks, x_blocks, axis),
            p_dask,
            x,
            dtype=np.float64).compute()
    else:
        x_ndarr = _to_numpy_ndarray(x)
        y = _ndpolyval(p_ndarr, x_ndarr, axis)

    attrs = {"p": p} if kwargs.get("return_info", False) else {}

    output = xr.DataArray(y, attrs=attrs)
    return output


def _ndpolyval(p: np.ndarray,
               x: np.ndarray,
               axis: int = 0,
               **kwargs) -> np.ndarray:
    if not isinstance(p, np.ndarray):
        raise TypeError("This function accepts only numpy.ndarray as p.")

    if not isinstance(x, np.ndarray):
        raise TypeError("This function accepts only numpy.ndarray as x.")

    axis = _check_axis(axis, p.ndim)

    if (x.ndim != 1) and (x.ndim != p.ndim):
        raise ValueError(
            "x has invalid number of dimension. x must be either 1 dimensionn.")

    if _isvector(x):
        x_original_size = x.size
        other_dims = np.asarray(p.shape)[np.arange(p.ndim) != axis]
        x = np.moveaxis(
            np.tile(x.reshape((-1,)), np.prod(other_dims)).reshape(
                (*other_dims, x_original_size)), -1, axis)

    else:
        if not (np.all(x.shape[:axis] == p.shape[:axis]) and np.all(
                x.shape[(axis + 1):x.ndim] == p.shape[(axis + 1):p.ndim])):
            raise ValueError("x has invalid shape.")

    y = np.zeros(x.shape)

    for i in range(p.shape[axis]):
        y += p.take([i], axis=axis) * np.power(x, p.shape[axis] - 1 - i)

    return y


def detrend(data: Iterable, deg=1, axis=0, **kwargs):
    """Estimates and removes the trend of the leftmost dimension from all grid
    points. This method, at the minimum, provides all the functionality that is
    provided by NCL's 'dtrend', 'dtrend_quadratic', 'dtrend_quadratic_msg_n',
    'dtrend_msg_n', 'dtrend_msg', 'dtrend_n'. However, this function is not
    limited to quadratic detrending and you could use higher polynomial degree
    as well.

    Parameters
    ----------
    data : :class:`array_like`
        a multi-dimensional numeric array

    deg : :class:`int`, optional
        a non-negative integer determining the degree of the polynomial to use for detrending. Default value is 1.

    axis : :class:`int`, optional
        the axis along which the data is detrended. Default value is 0.

    kwargs : :class:`dict`, optional
        providing further arguments to control the behavior of the method. It currently accepts the following parameter:

        return_info (:class:`bool`)
            If set to true, the fitted polynomial is returned as part of the attributes. Default value is `True`.

        missing_value (:class:`numeric`)
            A value that must be ignored. Default is NaN.

    Returns
    -------
    detrended_data : :class:`xarray.DataArray`
        Array containing the detrended data.

    Examples
    --------

        * Detrending a data:

        >>> from geocat.comp.polynomial import ndpolyfit
        >>> from geocat.comp.polynomial import detrend
        >>> # Creating synthetic data
        >>> x = np.linspace(-8*np.pi, 8 * np.pi, 33, dtype=np.float64)
        >>> y0 = 1.0 * x
        >>> y1 = np.sin(x)
        >>> y = y0 + y1
        >>> p = ndpolyfit(np.arange(x.size), y, deg=1)
        >>> y_trend = ndpolyval(p, np.arange(x.size))
        >>> y_detrended = detrend(y)
        >>> np.testing.assert_almost_equal(y_detrended + y_trend, y)


        * Detrending a multi-dimensional data:

        >>> # Creating synthetic data
        >>> x = np.linspace(-8*np.pi, 8 * np.pi, 33, dtype=np.float64)
        >>> y0 = 1.0 * x
        >>> y1 = np.sin(x)
        >>> y = np.tile((y0 + y1).reshape((1, -1, 1, 1)), (2, 1, 3, 4))
        >>> p = ndpolyfit(x, y, deg=1, axis=1)
        >>> y_trend = ndpolyval(p, x, axis=1)
        >>> y_detrended = detrend(y, x=x, axis=1)
        >>> np.testing.assert_almost_equal(y_detrended + y_trend, y)
    """
    if (int(deg) != deg) or (deg < 0):
        raise ValueError("deg must be non-negative integer value.")

    return_info = bool(kwargs.get("return_info", True))

    missing_value = _get_missing_value(data, kwargs)

    try:
        data_shape = data.shape if isinstance(
            data,
            (np.ndarray, xr.DataArray, da.Array)) else np.asarray(data).shape
    except BaseException:
        raise TypeError("Could not extract the shape of data")

    x = kwargs.get("x", None)
    if x is None:
        x = np.arange(data_shape[axis], dtype=np.float)
    elif not isinstance(x, np.ndarray):
        x = np.asarray(x).astype(np.float)

    x = x.reshape((-1,))

    p = ndpolyfit(x, data, deg, axis, missing_value=missing_value)

    data_fitted = ndpolyval(p, x, axis)

    data_detrended = xr.DataArray(data) - data_fitted

    if return_info:
        data_detrended.attrs["p"] = p

    return data_detrended



# move functions into geocat.comp namespace
from .climatology import anomaly, climatology, month_to_season
from .crop import (actual_saturation_vapor_pressure, max_daylight,
                   psychrometric_constant, saturation_vapor_pressure,
                   saturation_vapor_pressure_slope)
from .dewtemp import dewtemp
from .eofunc import eofunc, eofunc_eofs, eofunc_pcs, eofunc_ts
from .errors import (AttributeError, ChunkError, CoordinateError,
                     DimensionError, Error, MetaError)
from .fourier_filters import (fourier_band_block, fourier_band_pass,
                              fourier_filter, fourier_high_pass,
                              fourier_low_pass)
from .heat_index import heat_index
from .interp_hybrid_to_pressure import interp_hybrid_to_pressure
from .polynomial import detrend, ndpolyfit, ndpolyval
from .relhum import relhum, relhum_ice, relhum_water
from .skewt_params import get_skewt_vars, showalter_index

# bring all functions from geocat.f2py into the geocat.comp namespace
try:
    from geocat.f2py import *
except ImportError:
    pass



import warnings

import numpy as np
import xarray as xr

from .comp_util import _is_duck_array


def _nws_eqn(coeffs, temp, rel_hum):
    heatindex = coeffs[0] \
                + coeffs[1] * temp \
                + coeffs[2] * rel_hum \
                + coeffs[3] * temp * rel_hum \
                + coeffs[4] * temp ** 2 \
                + coeffs[5] * rel_hum ** 2 \
                + coeffs[6] * temp ** 2 * rel_hum \
                + coeffs[7] * temp * rel_hum ** 2 \
                + coeffs[8] * temp ** 2 * rel_hum ** 2

    return heatindex


def heat_index(temperature, relative_humidity, alternate_coeffs=False):
    """Compute the 'heat index' as calculated by the National Weather Service.

    The heat index calculation in this funtion is described at:
    https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml

    The 'Heat Index' is a measure of how hot weather "feels" to the body. The combination of temperature an humidity
    produce an "apparent temperature" or the temperature the body "feels". The returned values are for shady
    locations only. Exposure to full sunshine can increase heat index values by up to 15Â°F. Also, strong winds,
    particularly with very hot, dry air, can be extremely hazardous as the wind adds heat to the body

    The computation of the heat index is a refinement of a result obtained by multiple regression analysis carried
    out by Lans P. Rothfusz and described in a 1990 National Weather Service (NWS) Technical Attachment (SR 90-23).
    All values less that 40F/4.4C/277.65K are set to the ambient temperature.

    In practice, the Steadman formula is computed first and the result averaged with the temperature. If this heat
    index value is 80 degrees F or higher, the full regression equation along with any adjustment as described above
    is applied. If the ambient temperature is less the 40F (4.4C/277.65K), the heat index is set to to the ambient
    temperature.

    Parameters
    ----------
    temperature : numpy.ndarray, xr.DataArray, float
        temperature(s) in Fahrenheit

    relative_humidity : numpy.ndarray, xr.DataArray, float
        relative humidity as a percentage. Must be the same shape as
        temperature

    alternate_coeffs : Boolean, Optional
        flag to use alternate set of coefficients appropriate for
        temperatures from 70F to 115F and humidities between 0% and 80%

    Returns
    -------
    heatindex : numpy.ndarray, xr.DataArray
        Calculated heat index. Same shape as temperature

    Examples
    --------
    >>> import numpy as np
    >>> import geocat.comp
    >>> t = np.array([104, 100, 92])
    >>> rh = np.array([55, 65, 60])
    >>> hi = heat_index(t,rh)
    >>> hi
    array([137.36135724, 135.8679973 , 104.68441864])
    """

    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords

    # convert inputs to numpy arrays if necessary
    if not _is_duck_array(temperature):
        temperature = np.atleast_1d(temperature)
    if not _is_duck_array(relative_humidity):
        relative_humidity = np.atleast_1d(relative_humidity)

    # Input validation on relative humidity
    if any(relative_humidity.ravel() < 0) or any(
            relative_humidity.ravel() > 100):
        raise ValueError('heat_index: invalid values for relative humidity')

    # Check if relative humidity fractional
    if all(relative_humidity.ravel() < 1):
        warnings.warn("WARNING: rh must be %, not fractional; All rh are < 1")

    # Default coefficients for (t>=80F) and (40<gh<100)
    coeffs = [
        -42.379, 2.04901523, 10.14333127, -0.22475541, -0.00683783, -0.05481717,
        0.00122874, 0.00085282, -0.00000199
    ]
    crit = [80, 40, 100]  # [T_low [F], RH_low, RH_high]

    # Optional flag coefficients for (70F<t<115F) and (0<gh<80)
    # within 3F of default coeffs
    if alternate_coeffs:
        coeffs = [
            0.363445176, 0.988622465, 4.777114035, -0.114037667, -0.000850208,
            -0.020716198, 0.000687678, 0.000274954, 0.0
        ]
        crit = [70, 0, 80]  # [T_low [F], RH_low, RH_high]

    # NWS practice
    # average Steadman and t
    heatindex = (0.5 * (temperature + 61.0 + ((temperature - 68.0) * 1.2) +
                        (relative_humidity * 0.094)) + temperature) * 0.5

    # http://ehp.niehs.nih.gov/1206273/
    heatindex = np.where(temperature < 40, temperature, heatindex)

    # if all t values less than critical, return hi
    # otherwise perform calculation
    eqtype = 0
    if not all(temperature.ravel() < crit[0]):
        eqtype = 1

        heatindex = np.where(heatindex > crit[0],
                             _nws_eqn(coeffs, temperature, relative_humidity),
                             heatindex)

        # adjustments
        heatindex = np.where(
            np.logical_and(relative_humidity < 13,
                           np.logical_and(temperature > 80, temperature < 112)),
            heatindex - ((13 - relative_humidity) / 4) * np.sqrt(
                (17 - abs(temperature - 95)) / 17), heatindex)

        heatindex = np.where(
            np.logical_and(relative_humidity > 85,
                           np.logical_and(temperature > 80, temperature < 87)),
            heatindex + ((relative_humidity - 85.0) / 10.0) *
            ((87.0 - temperature) / 5.0), heatindex)

    # reformat output for xarray if necessary
    if x_out:
        heatindex = xr.DataArray(heatindex, coords=save_coords, dims=save_dims)
        heatindex.attrs['long_name'] = "heat index: NWS"
        heatindex.attrs['units'] = "F"
        heatindex.attrs[
            'www'] = "https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml"
        heatindex.attrs['info'] = "appropriate for shady locations with no wind"

        if eqtype == 1:
            heatindex.attrs[
                'tag'] = "NCL: heat_index_nws; (Steadman+t)*0.5 and Rothfusz"
        else:
            heatindex.attrs['tag'] = "NCL: heat_index_nws; (Steadman+t)*0.5"

    return heatindex



import numpy as np
import xarray as xr


def dewtemp(temperature, relative_humidity):
    """This function calculates the dew point temperature given temperature and
    relative humidity using equations from John Dutton's "Ceaseless Wind" (pp
    273-274)

    Args:

        temperature (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
            Temperature in K

        relative_humidity (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
            Relative humidity. Must be the same dimensions as temperature


    Returns:

        dew_pnt_temp (:class:`numpy.ndarray` or :obj:`float`):
            Dewpoint temperature in Kelvin. Same size as input variable temperature
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    relative_humidity = np.asarray(relative_humidity)

    # make sure the input arrays are of the same size
    if np.shape(temperature) != np.shape(relative_humidity):
        raise ValueError(
            f"dewtemp_trh: dimensions of temperature, {np.shape(temperature)}, and relative_humidity, "
            f"{np.shape(relative_humidity)}, do not match")

    # Call mapblocks to run function
    dew_pnt_temp = _dewtemp(temperature, relative_humidity)

    # output as xarray if input as xarray
    if x_out:
        dew_pnt_temp = xr.DataArray(data=dew_pnt_temp,
                                    coords=save_coords,
                                    dims=save_dims,
                                    attrs=save_attrs)

    return dew_pnt_temp


def _dewtemp(tk, rh):
    """This function calculates the dew point temperature given temperature and
    relative humidity using equations from John Dutton's "Ceaseless Wind" (pp
    273-274)

    Args:

        tk (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
            Temperature in K

        rh (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
            Relative humidity. Must be the same dimensions as temperature


    Returns:

        tdk (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
            Dewpoint temperature in Kelvin. Same size as input variable temperature
    """

    gc = 461.5  # gas constant for water vapor [j/{kg-k}]
    gcx = gc / (1000 * 4.186)  # [cal/{g-k}]

    lhv = (597.3 - 0.57 * (tk - 273)) / gcx
    tdk = tk * lhv / (lhv - tk * np.log(rh * 0.01))

    return tdk



class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class AttributeError(Error):
    """Exception raised when the arguments of GeoCAT-comp functions argument
    has a mismatch of attributes with other arguments."""
    pass


class ChunkError(Error):
    """Exception raised when a Dask array is chunked in a way that is
    incompatible with an f2py function."""
    pass


class CoordinateError(Error):
    """Exception raised when a GeoCAT-comp function is passed a NumPy array as
    an argument without a required coordinate array being passed separately."""
    pass


class DimensionError(Error):
    """Exception raised when the arguments of GeoCAT-comp functions argument
    has a mismatch of the necessary dimensionality."""
    pass


class MetaError(Error):
    """Exception raised when the support for the retention of metadata is not
    supported."""
    pass



import cf_xarray
import metpy.interpolate
import numpy as np
import xarray as xr

__pres_lev_mandatory__ = np.array([
    1000, 925, 850, 700, 500, 400, 300, 250, 200, 150, 100, 70, 50, 30, 20, 10,
    7, 5, 3, 2, 1
]).astype(np.float32)  # Mandatory pressure levels (mb)
__pres_lev_mandatory__ = __pres_lev_mandatory__ * 100.0  # Convert mb to Pa


def _pressure_from_hybrid(psfc, hya, hyb, p0=100000.):
    """Calculate pressure at the hybrid levels."""
    # This will be in Pa
    return hya * p0 + hyb * psfc


def interp_hybrid_to_pressure(data,
                              ps,
                              hyam,
                              hybm,
                              p0=100000.,
                              new_levels=__pres_lev_mandatory__,
                              lev_dim=None,
                              method='linear'):
    """Interpolate data from hybrid-sigma levels to isobaric levels.

    Notes
    -----
    ACKNOWLEDGEMENT: We'd like to thank to Brian Medeiros (https://github.com/brianpm), Matthew Long
    (https://github.com/matt-long), and Deepak Cherian (https://github.com/dcherian) at NCAR for their
    great contributions since the code implemented here is mostly based on their work.

    Parameters
    ----------
    data : xarray.DataArray
        Multidimensional data array, which holds hybrid-sigma levels and has a `lev_dim` coordinate.

    ps : xarray.DataArray
        A multi-dimensional array of surface pressures (Pa), same time/space shape as data.

    hyam, hybm : xarray.DataArray
        One-dimensional arrays containing the hybrid A and B coefficients. Must have the same
        dimension size as the `lev_dim` dimension of data.

    p0 :
        Scalar numeric value equal to surface reference pressure (Pa).

    new_levels : np.ndarray
        A one-dimensional array of output pressure levels (Pa). If not given, the mandatory
        list of 21 pressure levels is used.

    lev_dim : str
        String that is the name of level dimension in data. Defaults to "lev".

    method : str
        String that is the interpolation method; can be either "linear" or "log". Defaults to "linear".
    """

    # Determine the level dimension and then the interpolation axis
    if lev_dim is None:
        try:
            lev_dim = data.cf["vertical"].name
        except Exception:
            raise ValueError(
                "Unable to determine vertical dimension name. Please specify the name via `lev_dim` argument.'"
            )

    interp_axis = data.dims.index(lev_dim)

    # Calculate pressure levels at the hybrid levels
    pressure = _pressure_from_hybrid(ps, hyam, hybm, p0)  # Pa

    # Define interpolation function
    if method == 'linear':
        func_interpolate = metpy.interpolate.interpolate_1d
    elif method == 'log':
        func_interpolate = metpy.interpolate.log_interpolate_1d
    else:
        raise ValueError(f'Unknown interpolation method: {method}. '
                         f'Supported methods are: "log" and "linear".')

    def _vertical_remap(data, pressure):
        """Define interpolation function."""

        return func_interpolate(new_levels, pressure, data, axis=interp_axis)

    # Apply vertical interpolation
    # Apply Dask parallelization with xarray.apply_ufunc
    output = xr.apply_ufunc(
        _vertical_remap,
        data,
        pressure,
        exclude_dims=set((lev_dim,)),  # Set dimensions allowed to change size
        input_core_dims=[[lev_dim], [lev_dim]],  # Set core dimensions
        output_core_dims=[["plev"]],  # Specify output dimensions
        vectorize=True,  # loop over non-core dims
        dask="parallelized",  # Dask parallelization
        output_dtypes=[data.dtype],
        dask_gufunc_kwargs={"output_sizes": {
            "plev": len(new_levels)
        }},
    )

    # Set output dims and coords
    dims = ["plev"
           ] + [data.dims[i] for i in range(data.ndim) if i != interp_axis]

    coords = {}
    for (k, v) in data.coords.items():
        if k != lev_dim:
            coords.update({k: v})
        else:
            coords.update({"plev": new_levels})

    output = output.transpose(*dims).assign_coords(coords)

    return output



import numpy as np
import xarray as xr


def relhum(temperature, mixing_ratio, pressure):
    """This function calculates the relative humidity given temperature, mixing
    ratio, and pressure.

    "Improved Magnus' Form Approx. of Saturation Vapor pressure"
    Oleg A. Alduchov and Robert E. Eskridge
    http://www.osti.gov/scitech/servlets/purl/548871/
    https://doi.org/10.2172/548871

        Args:

            temperature (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
                Temperature in Kelvin

            mixing_ratio (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
                Mixing ratio in kg/kg. Must have the same dimensions as temperature

            pressure (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
                Pressure in Pa. Must have the same dimensions as temperature

        Returns:

            relative_humidity (:class:`numpy.ndarray` or :class:`xr.DataArray`):
                Relative humidity. Will have the same dimensions as temperature
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    mixing_ratio = np.asarray(mixing_ratio)
    pressure = np.asarray(pressure)

    # ensure all inputs same size
    if np.shape(temperature) != np.shape(mixing_ratio) or np.shape(
            temperature) != np.shape(pressure):
        raise ValueError(f"relhum: dimensions of inputs are not the same")

    relative_humidity = _relhum(temperature, mixing_ratio, pressure)

    # output as xarray if input as xarray
    if x_out:
        relative_humidity = xr.DataArray(data=relative_humidity,
                                         coords=save_coords,
                                         dims=save_dims,
                                         attrs=save_attrs)

    return relative_humidity


def relhum_water(temperature, mixing_ratio, pressure):
    """Calculates relative humidity with respect to water, given temperature,
    mixing ratio, and pressure.

     Definition of mixing ratio if,
     es  - is the saturation mixing ratio
     ep  - is the ratio of the molecular weights of water vapor to dry air
     p   - is the atmospheric pressure
     rh  - is the relative humidity (given as a percent)

     rh =  100*  q / ( (ep*es)/(p-es) )

    Args:

         temperature (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Temperature in Kelvin

         mixing_ratio (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Mixing ratio in kg/kg. Must have the same dimensions as temperature

         pressure (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Pressure in Pa. Must have the same dimensions as temperature

     Returns:

         relative_humidity (:class:`numpy.ndarray`):
             Relative humidity. Will have the same dimensions as temperature
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    mixing_ratio = np.asarray(mixing_ratio)
    pressure = np.asarray(pressure)

    # ensure all inputs same size
    if np.shape(temperature) != np.shape(mixing_ratio) or np.shape(
            temperature) != np.shape(pressure):
        raise ValueError(f"relhum_water: dimensions of inputs are not the same")

    relative_humidity = _relhum_water(temperature, mixing_ratio, pressure)

    # output as xarray if input as xarray
    if x_out:
        relative_humidity = xr.DataArray(data=relative_humidity,
                                         coords=save_coords,
                                         dims=save_dims,
                                         attrs=save_attrs)

    return relative_humidity


def relhum_ice(temperature, mixing_ratio, pressure):
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

     "Improved Magnus' Form Approx. of Saturation Vapor pressure"
     Oleg A. Alduchov and Robert E. Eskridge
     http://www.osti.gov/scitech/servlets/purl/548871/
     https://doi.org/10.2172/548871

    Args:

         temperature (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Temperature in Kelvin

         mixing_ratio (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Mixing ratio in kg/kg. Must have the same dimensions as temperature

         pressure (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Pressure in Pa. Must have the same dimensions as temperature

     Returns:

         relative_humidity (:class:`numpy.ndarray`):
             Relative humidity. Will have the same dimensions as temperature
    """

    # If xarray input, pull data and store metadata
    x_out = False
    if isinstance(temperature, xr.DataArray):
        x_out = True
        save_dims = temperature.dims
        save_coords = temperature.coords
        save_attrs = temperature.attrs

    # ensure in numpy array for function call
    temperature = np.asarray(temperature)
    mixing_ratio = np.asarray(mixing_ratio)
    pressure = np.asarray(pressure)

    # ensure all inputs same size
    if np.shape(temperature) != np.shape(mixing_ratio) or np.shape(
            temperature) != np.shape(pressure):
        raise ValueError(f"relhum_ice: dimensions of inputs are not the same")

    relative_humidity = _relhum_ice(temperature, mixing_ratio, pressure)

    # output as xarray if input as xarray
    if x_out:
        relative_humidity = xr.DataArray(data=relative_humidity,
                                         coords=save_coords,
                                         dims=save_dims,
                                         attrs=save_attrs)

    return relative_humidity


def _relhum(t, w, p):
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

     "Improved Magnus' Form Approx. of Saturation Vapor pressure"
     Oleg A. Alduchov and Robert E. Eskridge
     http://www.osti.gov/scitech/servlets/purl/548871/
     https://doi.org/10.2172/548871

    Args:

         t (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Temperature in Kelvin

         w (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
             Mixing ratio in kg/kg. Must have the same dimensions as temperature

         p (:class:`numpy.ndarray`, :class:`xr.DataArray`, :obj:`list`, or :obj:`float`):
             Pressure in Pa. Must have the same dimensions as temperature

     Returns:

         rh (:class:`numpy.ndarray):
             Relative humidity. Will have the same dimensions as temperature
    """

    table = np.asarray([
        0.01403, 0.01719, 0.02101, 0.02561, 0.03117, 0.03784, 0.04584, 0.05542,
        0.06685, 0.08049, 0.09672, 0.1160, 0.1388, 0.1658, 0.1977, 0.2353,
        0.2796, 0.3316, 0.3925, 0.4638, 0.5472, 0.6444, 0.7577, 0.8894, 1.042,
        1.220, 1.425, 1.662, 1.936, 2.252, 2.615, 3.032, 3.511, 4.060, 4.688,
        5.406, 6.225, 7.159, 8.223, 9.432, 10.80, 12.36, 14.13, 16.12, 18.38,
        20.92, 23.80, 27.03, 30.67, 34.76, 39.35, 44.49, 50.26, 56.71, 63.93,
        71.98, 80.97, 90.98, 102.1, 114.5, 128.3, 143.6, 160.6, 179.4, 200.2,
        223.3, 248.8, 276.9, 307.9, 342.1, 379.8, 421.3, 466.9, 517.0, 572.0,
        632.3, 698.5, 770.9, 850.2, 937.0, 1032.0, 1146.6, 1272.0, 1408.1,
        1556.7, 1716.9, 1890.3, 2077.6, 2279.6, 2496.7, 2729.8, 2980.0, 3247.8,
        3534.1, 3839.8, 4164.8, 4510.5, 4876.9, 5265.1, 5675.2, 6107.8, 6566.2,
        7054.7, 7575.3, 8129.4, 8719.2, 9346.50, 10013.0, 10722.0, 11474.0,
        12272.0, 13119.0, 14017.0, 14969.0, 15977.0, 17044.0, 18173.0, 19367.0,
        20630.0, 21964.0, 23373.0, 24861.0, 26430.0, 28086.0, 29831.0, 31671.0,
        33608.0, 35649.0, 37796.0, 40055.0, 42430.0, 44927.0, 47551.0, 50307.0,
        53200.0, 56236.0, 59422.0, 62762.0, 66264.0, 69934.0, 73777.0, 77802.0,
        82015.0, 86423.0, 91034.0, 95855.0, 100890.0, 106160.0, 111660.0,
        117400.0, 123400.0, 129650.0, 136170.0, 142980.0, 150070.0, 157460.0,
        165160.0, 173180.0, 181530.0, 190220.0, 199260.0, 208670.0, 218450.0,
        228610.0, 239180.0, 250160.0, 261560.0, 273400.0, 285700.0, 298450.0,
        311690.0, 325420.0, 339650.0, 354410.0, 369710.0, 385560.0, 401980.0,
        418980.0, 436590.0, 454810.0, 473670.0, 493170.0, 513350.0, 534220.0,
        555800.0, 578090.0, 601130.0, 624940.0, 649530.0, 674920.0, 701130.0,
        728190.0, 756110.0, 784920.0, 814630.0, 845280.0, 876880.0, 909450.0,
        943020.0, 977610.0, 1013250.0, 1049940.0, 1087740.0, 1087740.
    ])

    maxtemp = 375.16
    mintemp = 173.16

    # replace values of t above and below max and min values for temperature
    t = np.clip(t, mintemp, maxtemp)

    it = (t - mintemp).astype(int)
    t2 = mintemp + it

    es = (t2 + 1 - t) * table[it] + (t - t2) * table[it + 1]
    es = es * 0.1

    rh = (w * (p - 0.378 * es) / (0.622 * es)) * 100

    # if any value is below 0.0001, set to 0.0001
    rh = np.clip(rh, 0.0001, None)

    return rh


def _relhum_ice(t, w, p):
    """Calculates relative humidity with respect to ice, given temperature,
    mixing ratio, and pressure.

     "Improved Magnus' Form Approx. of Saturation Vapor pressure"
     Oleg A. Alduchov and Robert E. Eskridge
     http://www.osti.gov/scitech/servlets/purl/548871/
     https://doi.org/10.2172/548871

    Args:

         t (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Temperature in Kelvin

         w (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Mixing ratio in kg/kg. Must have the same dimensions as temperature

         p (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Pressure in Pa. Must have the same dimensions as temperature

     Returns:

         rh (:class:`numpy.ndarray`):
             Relative humidity. Will have the same dimensions as temperature
    """

    # Define data variables

    t0 = 273.15
    ep = 0.622
    onemep = 0.378
    es0 = 6.1128
    a = 22.571
    b = 273.71

    est = es0 * np.exp((a * (t - t0)) / ((t - t0) + b))
    qst = (ep * est) / ((p * 0.01) - onemep * est)

    rh = 100 * (w / qst)

    return rh


def _relhum_water(t, w, p):
    """Calculates relative humidity with respect to water, given temperature,
    mixing ratio, and pressure.

     Definition of mixing ratio if,
     es  - is the saturation mixing ratio
     ep  - is the ratio of the molecular weights of water vapor to dry air
     p   - is the atmospheric pressure
     rh  - is the relative humidity (given as a percent)

     rh =  100*  q / ( (ep*es)/(p-es) )

    Args:

         t (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Temperature in Kelvin

         w (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Mixing ratio in kg/kg. Must have the same dimensions as temperature

         p (:class:`numpy.ndarray`, :obj:`list`, or :obj:`float`):
             Pressure in Pa. Must have the same dimensions as temperature

     Returns:

         rh (:class:`numpy.ndarray`):
             Relative humidity. Will have the same dimensions as temperature
    """

    # Define data variables

    t0 = 273.15
    ep = 0.622
    onemep = 0.378
    es0 = 6.1128
    a = 17.269
    b = 35.86

    est = es0 * np.exp((a * (t - t0)) / (t - b))
    qst = (ep * est) / ((p * 0.01) - onemep * est)

    rh = 100 * (w / qst)

    return rh
